@@ -5,24 +5,20 @@ use std::borrow::Cow;
 use indoc::formatdoc;
 use js_int::UInt;
 use ruma_events_macros::EventContent;
-use ruma_identifiers::MxcUri;
 #[cfg(feature = "unstable-pre-spec")]
 use ruma_identifiers::{DeviceIdBox, UserId};
+use ruma_identifiers::{EventId, MxcUri};
 use ruma_serde::StringEnum;
 use serde::{de::DeserializeOwned, Deserialize, Serialize};
 use serde_json::Value as JsonValue;
 
-#[cfg(feature = "unstable-pre-spec")]
-use super::relationships::{Annotation, Reference, RelationJsonRepr, Replacement};
-use super::{relationships::RelatesToJsonRepr, EncryptedFile, ImageInfo, ThumbnailInfo};
+use super::{EncryptedFile, ImageInfo, ThumbnailInfo};
 #[cfg(feature = "unstable-pre-spec")]
 use crate::key::verification::VerificationMethod;
 
-// FIXME: Do we want to keep re-exporting this?
-pub use super::relationships::InReplyTo;
-
 mod content_serde;
 pub mod feedback;
+mod relation_serde;
 
 type JsonObject = serde_json::Map<String, JsonValue>;
 
@@ -42,29 +38,17 @@ pub struct MessageEventContent {
     #[serde(flatten)]
     pub msgtype: MessageType,
 
-    /// Information about related messages for
-    /// [rich replies](https://matrix.org/docs/spec/client_server/r0.6.1#rich-replies).
-    #[serde(rename = "m.relates_to", skip_serializing_if = "Option::is_none")]
-    pub relates_to: Option<Relation>,
-
-    /// New content of an edited message.
+    /// Information about related messages for [rich replies].
     ///
-    /// This should only be set if `relates_to` is `Some(Relation::Replacement(_))`.
-    #[cfg(feature = "unstable-pre-spec")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "unstable-pre-spec")))]
-    #[serde(rename = "m.new_content", skip_serializing_if = "Option::is_none")]
-    pub new_content: Option<Box<MessageEventContent>>,
+    /// [rich replies]: https://matrix.org/docs/spec/client_server/r0.6.1#rich-replies
+    #[serde(flatten, with = "relation_serde", skip_serializing_if = "Option::is_none")]
+    pub relates_to: Option<Relation>,
 }
 
 impl MessageEventContent {
     /// Create a `MessageEventContent` with the given `MessageType`.
     pub fn new(msgtype: MessageType) -> Self {
-        Self {
-            msgtype,
-            relates_to: None,
-            #[cfg(feature = "unstable-pre-spec")]
-            new_content: None,
-        }
+        Self { msgtype, relates_to: None }
     }
 
     /// A constructor to create a plain text message.
@@ -276,67 +260,57 @@ impl From<MessageType> for MessageEventContent {
     }
 }
 
-/// Enum modeling the different ways relationships can be expressed in a
-/// `m.relates_to` field of an `m.room.message` event.
-#[derive(Clone, Debug, Deserialize, Serialize)]
-#[serde(from = "RelatesToJsonRepr", into = "RelatesToJsonRepr")]
+/// Message vent relationship.
+///
+/// Currently used for replies and editing (message replacement).
+#[derive(Clone, Debug)]
+#[cfg_attr(not(feature = "unstable-exhaustive-types"), non_exhaustive)]
 pub enum Relation {
-    /// A reference to another event.
-    #[cfg(feature = "unstable-pre-spec")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "unstable-pre-spec")))]
-    Reference(Reference),
-
-    /// An annotation to an event.
-    #[cfg(feature = "unstable-pre-spec")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "unstable-pre-spec")))]
-    Annotation(Annotation),
-
-    /// An event that replaces another event.
-    #[cfg(feature = "unstable-pre-spec")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "unstable-pre-spec")))]
-    Replacement(Replacement),
-
-    /// An `m.in_reply_to` relation indicating that the event is a reply to
-    /// another event.
+    /// An `m.in_reply_to` relation indicating that the event is a reply to another event.
     Reply {
         /// Information about another message being replied to.
         in_reply_to: InReplyTo,
     },
 
-    /// Custom, unsupported relation.
-    #[doc(hidden)]
-    _Custom(JsonValue),
+    /// An event that replaces another event.
+    #[cfg(feature = "unstable-pre-spec")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "unstable-pre-spec")))]
+    Replacement(Replacement),
 }
 
-impl From<Relation> for RelatesToJsonRepr {
-    fn from(value: Relation) -> Self {
-        match value {
-            #[cfg(feature = "unstable-pre-spec")]
-            Relation::Annotation(r) => RelatesToJsonRepr::Relation(RelationJsonRepr::Annotation(r)),
-            #[cfg(feature = "unstable-pre-spec")]
-            Relation::Reference(r) => RelatesToJsonRepr::Relation(RelationJsonRepr::Reference(r)),
-            #[cfg(feature = "unstable-pre-spec")]
-            Relation::Replacement(r) => {
-                RelatesToJsonRepr::Relation(RelationJsonRepr::Replacement(r))
-            }
-            Relation::Reply { in_reply_to } => RelatesToJsonRepr::Reply { in_reply_to },
-            Relation::_Custom(c) => RelatesToJsonRepr::Custom(c),
-        }
+/// Information about the event a "rich reply" is replying to.
+#[derive(Clone, Debug, Deserialize, Serialize)]
+#[cfg_attr(not(feature = "unstable-exhaustive-types"), non_exhaustive)]
+pub struct InReplyTo {
+    /// The event being replied to.
+    pub event_id: EventId,
+}
+
+impl InReplyTo {
+    /// Creates a new `InReplyTo` with the given event ID.
+    pub fn new(event_id: EventId) -> Self {
+        Self { event_id }
     }
 }
 
-impl From<RelatesToJsonRepr> for Relation {
-    fn from(value: RelatesToJsonRepr) -> Self {
-        match value {
-            #[cfg(feature = "unstable-pre-spec")]
-            RelatesToJsonRepr::Relation(r) => match r {
-                RelationJsonRepr::Annotation(a) => Self::Annotation(a),
-                RelationJsonRepr::Reference(r) => Self::Reference(r),
-                RelationJsonRepr::Replacement(r) => Self::Replacement(r),
-            },
-            RelatesToJsonRepr::Reply { in_reply_to } => Self::Reply { in_reply_to },
-            RelatesToJsonRepr::Custom(v) => Self::_Custom(v),
-        }
+/// The event this relation belongs to replaces another event.
+#[derive(Clone, Debug)]
+#[cfg(feature = "unstable-pre-spec")]
+#[cfg_attr(docsrs, doc(cfg(feature = "unstable-pre-spec")))]
+#[cfg_attr(not(feature = "unstable-exhaustive-types"), non_exhaustive)]
+pub struct Replacement {
+    /// The ID of the event being replacing.
+    pub event_id: EventId,
+
+    /// New content.
+    pub new_content: Box<MessageEventContent>,
+}
+
+#[cfg(feature = "unstable-pre-spec")]
+impl Replacement {
+    /// Creates a new `Replacement` with the given event ID and new content.
+    pub fn new(event_id: EventId, new_content: Box<MessageEventContent>) -> Self {
+        Self { event_id, new_content }
     }
 }
 
@@ -427,13 +401,13 @@ impl EmoteMessageEventContent {
     }
 
     /// A convenience constructor to create a markdown emote.
+    ///
+    /// Returns an html emote message if some markdown formatting was detected, otherwise returns a
+    /// plain-text emote.
     #[cfg(feature = "markdown")]
     #[cfg_attr(docsrs, doc(cfg(feature = "markdown")))]
-    pub fn markdown(body: impl Into<String>) -> Self {
-        let body = body.into();
-        let mut html_body = String::new();
-        pulldown_cmark::html::push_html(&mut html_body, pulldown_cmark::Parser::new(&body));
-        Self::html(body, html_body)
+    pub fn markdown(body: impl AsRef<str> + Into<String>) -> Self {
+        Self { formatted: FormattedBody::markdown(&body), ..Self::plain(body) }
     }
 }
 
@@ -629,6 +603,16 @@ impl NoticeMessageEventContent {
     pub fn html(body: impl Into<String>, html_body: impl Into<String>) -> Self {
         Self { formatted: Some(FormattedBody::html(html_body)), ..Self::plain(body) }
     }
+
+    /// A convenience constructor to create a markdown notice.
+    ///
+    /// Returns an html notice if some markdown formatting was detected, otherwise returns a plain
+    /// text notice.
+    #[cfg(feature = "markdown")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "markdown")))]
+    pub fn markdown(body: impl AsRef<str> + Into<String>) -> Self {
+        Self { formatted: FormattedBody::markdown(&body), ..Self::plain(body) }
+    }
 }
 
 /// The payload for a server notice message.
@@ -711,6 +695,7 @@ impl MessageFormat {
 /// Common message event content fields for message types that have separate plain-text and
 /// formatted representations.
 #[derive(Clone, Debug, Deserialize, Serialize)]
+#[allow(clippy::exhaustive_structs)]
 pub struct FormattedBody {
     /// The format used in the `formatted_body`.
     pub format: MessageFormat,
@@ -725,6 +710,20 @@ impl FormattedBody {
     pub fn html(body: impl Into<String>) -> Self {
         Self { format: MessageFormat::Html, body: body.into() }
     }
+
+    /// Creates a new HTML-formatted message body by parsing the markdown in `body`.
+    ///
+    /// Returns `None` if no markdown formatting was found.
+    #[cfg(feature = "markdown")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "markdown")))]
+    pub fn markdown(body: impl AsRef<str>) -> Option<Self> {
+        let body = body.as_ref();
+        let mut html_body = String::new();
+
+        pulldown_cmark::html::push_html(&mut html_body, pulldown_cmark::Parser::new(body));
+
+        (html_body != format!("<p>{}</p>\n", body)).then(|| Self::html(html_body))
+    }
 }
 
 /// The payload for a text message.
@@ -752,13 +751,13 @@ impl TextMessageEventContent {
     }
 
     /// A convenience constructor to create a markdown message.
+    ///
+    /// Returns an html message if some markdown formatting was detected, otherwise returns a plain
+    /// text message.
     #[cfg(feature = "markdown")]
     #[cfg_attr(docsrs, doc(cfg(feature = "markdown")))]
-    pub fn markdown(body: impl Into<String>) -> Self {
-        let body = body.into();
-        let mut html_body = String::new();
-        pulldown_cmark::html::push_html(&mut html_body, pulldown_cmark::Parser::new(&body));
-        Self::html(body, html_body)
+    pub fn markdown(body: impl AsRef<str> + Into<String>) -> Self {
+        Self { formatted: FormattedBody::markdown(&body), ..Self::plain(body) }
     }
 
     /// A convenience constructor to create a plain text message.
@@ -1165,3 +1164,35 @@ fn formatted_or_plain_body<'a>(formatted: &'a Option<FormattedBody>, body: &'a s
         body
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use matches::assert_matches;
+    use ruma_identifiers::event_id;
+    use serde_json::{from_value as from_json_value, json};
+
+    use super::{InReplyTo, MessageEventContent, MessageType, Relation};
+
+    #[test]
+    fn deserialize_reply() {
+        let ev_id = event_id!("$1598361704261elfgc:localhost");
+
+        let json = json!({
+            "msgtype": "m.text",
+            "body": "<text msg>",
+            "m.relates_to": {
+                "m.in_reply_to": {
+                    "event_id": ev_id,
+                },
+            },
+        });
+
+        assert_matches!(
+            from_json_value::<MessageEventContent>(json).unwrap(),
+            MessageEventContent {
+                msgtype: MessageType::Text(_),
+                relates_to: Some(Relation::Reply { in_reply_to: InReplyTo { event_id } }),
+            } if event_id == ev_id
+        );
+    }
+}
